"""choose between remote config-vars, as defined in an heroku app."""
from functools import partial
import os
from typing import List, Optional

import click
import httpx
import questionary
from toolz import assoc, curry, dissoc, first, get_in, groupby

from chenv import console
from chenv.cli import cli
from chenv.models.output import Output

fatal = partial(console.fatal, __name__)


def _api_get(uri: str) -> httpx.Response:
    api_key = console.get_env_or_prompt(__name__, "HEROKU_API_KEY")
    return httpx.get(
        os.path.join("https://api.heroku.com", uri),
        headers=dict(
            Accept="application/vnd.heroku+json; version=3",
            Authorization=f"Bearer {api_key}",
        ),
    )


def _prompt_team_apps(default_team: Optional[str], apps: List[dict]) -> List[dict]:
    apps_by_team = groupby(curry(get_in, ["team", "name"]), apps)
    normalized_apps_by_team = dissoc(
        assoc(apps_by_team, "personal", apps_by_team[None]), None
    )

    if default_team and default_team in normalized_apps_by_team:
        team = default_team
    else:
        click.echo(
            f"""Teams found in account: {", ".join(sorted(
                click.style(team, fg="magenta")
                for team in normalized_apps_by_team))}"""
        )
        team = questionary.autocomplete(
            message="Choose team:", choices=normalized_apps_by_team
        ).ask()

    return normalized_apps_by_team[team]


def _prompt_app(default_app: Optional[str], apps: List[dict]) -> dict:
    apps_by_name = groupby("name", apps)
    if default_app and default_app in apps_by_name:
        name = default_app
    else:
        name = questionary.autocomplete(
            message="Choose app:", choices=apps_by_name
        ).ask()

    return first(apps_by_name[name])


def _choose_app(default_team: Optional[str], default_app: Optional[str]) -> dict:
    response = _api_get("apps")
    team_apps = _prompt_team_apps(default_team, response.json())
    return _prompt_app(default_app, team_apps)


def _get_config_vars(app: dict) -> dict:
    response = _api_get(f'apps/{app["id"]}/config-vars')
    return response.json()


@cli.command(help="choose between remote config-vars, as defined in an heroku app")
@click.option("--team", "-t", help="Heroku team name")
@click.option("--app", "-a", help="Heroku app name in the team")
def heroku(team: Optional[str], app: Optional[str]) -> Output:
    """Choose between remote config-vars, as defined in an heroku app."""
    chosen_app = _choose_app(team, app)
    vars = _get_config_vars(chosen_app)
    return Output(variables=vars, file_suffix=chosen_app["name"])



"""user inteaction utilities."""
from enum import Enum
import os
from typing import NoReturn

import click
import questionary

from chenv import settings


class Color(Enum):
    """Colors for output.

    docs: https://click.palletsprojects.com/en/7.x/api/#click.style
    """

    BLACK = "black"
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"
    MAGENTA = "magenta"
    CYAN = "cyan"
    WHITE = "white"
    BRIGHT_BLACK = "bright_black"
    BRIGHT_RED = "bright_red"
    BRIGHT_GREEN = "bright_green"
    BRIGHT_YELLOW = "bright_yellow"
    BRIGHT_BLUE = "bright_blue"
    BRIGHT_MAGENTA = "bright_magenta"
    BRIGHT_CYAN = "bright_cyan"
    BRIGHT_WHITE = "BRIGHT_WHITE"


def as_link(
    target: str,
    target_color: Color,
    source: str = settings.DOTENV,
    source_color: Color = Color.BRIGHT_BLACK,
) -> str:
    """Formart symlink relation, standardized, in color."""
    return (
        f"{click.style(source, fg=source_color.value)}"
        f" → {click.style(target, fg=target_color.value)}"
    )


def contextify(context: str, context_fg: Color, message: str) -> str:
    """Add context prefix to messages."""
    return f"{click.style(context, fg=context_fg.value)} - {message}"


def info(context: str, message: str) -> None:
    """Echo info colored messages with context."""
    click.echo(contextify(context, Color.CYAN, message))


def error(context: str, message: str) -> None:
    """Echo error colored messages with context."""
    click.echo(contextify(context, Color.RED, message))


def fatal(context: str, message: str, exit_code: int) -> NoReturn:
    """Exit the program after echoing an error colored messages with context."""
    error(
        context, f"({click.style(str(exit_code), fg=Color.YELLOW.value)}) - {message}"
    )
    if 0 < exit_code < 126:
        exit(exit_code)

    raise ValueError("`exit_code` for `fatal` must be in range [1, 126).")


def get_env_or_prompt(context: str, key: str) -> str:
    """Get or promt user for chenv-specific environment variable."""
    try:
        return os.environ[key]
    except KeyError:
        pass

    ephasized_key = click.style(key, fg=Color.YELLOW.value)
    info(context, f"Environment variable {ephasized_key} is not defined.")
    value = questionary.text("Define new value:", validate=bool).ask()
    if not value:
        raise ValueError(f"Cannot set value `{value}` to `{key}`")

    settings.add(key, value)
    settings.load()
    return value



"""Command-line interface."""
from typing import Optional

import click
from click.core import Context

from chenv import fs, settings, transform
from chenv.console import as_link, Color, fatal
from chenv.models.output import Output, Target
from . import __version__

settings.load()


@click.group(invoke_without_command=True)
@click.pass_context
@click.version_option(version=__version__)
def cli(ctx: Context) -> None:
    """chenv. modern local environment management."""
    click.secho("chenv. modern local environment management", fg=Color.GREEN.value)
    if ctx.invoked_subcommand is None:
        import chenv.subcommands

        ctx.invoke(chenv.subcommands.local.collect)


@cli.resultcallback()
def _pipeline(result: Optional[Target]) -> None:
    try:
        if result is None or result.file_suffix is None:
            return

        if isinstance(result, Output):
            transformed_output: Output = transform.apply(result)
            fs.assign_env(transformed_output)
        else:
            fs.force_link(result.filename)
        click.echo(as_link(result.filename, Color.BRIGHT_GREEN))
    except Exception as e:
        fatal(__name__, f"{e.__class__.__name__} - {str(e)}", 1)



"""filesystem utilities."""
import os
from typing import Generator

from chenv.models.output import Output


def append(file_path: str, *line: str) -> None:
    """Append `line`(s) to the file at `file_path`."""
    with open(file_path, "a") as f:
        f.writelines(line)


def assign_env(output: Output) -> None:
    """Dumps output as f'.env.{file_suffix}' file, then links .env to it."""
    dump(output.filename, output.body)
    force_link(output.filename)


def dump(filename: str, body: str) -> None:
    """Writes body to the file at `file_path`."""
    with open(filename, "w") as f:
        f.write(body)


def force_link(source_path: str) -> None:
    """Symlinks source_path to `.env`, forcibly."""
    try:
        os.remove(".env")
    except FileNotFoundError:
        pass

    os.symlink(source_path, ".env")


def load_lines(file_path: str) -> Generator[str, None, None]:
    """Creates a generator for lines in `file_path`."""
    try:
        with open(file_path) as f:
            while (line := f.readline()) :
                if (line := line.strip()) :
                    yield line
    except FileNotFoundError:
        return



"""show current link of the `.env` file."""
import os

import click

from chenv import settings
from chenv.cli import cli
from chenv.console import as_link, Color


@cli.command(help="show current link of the `.env` file.")
def which() -> None:
    """Show current link of the `.env` file."""
    try:
        target = os.readlink(settings.DOTENV)
        if os.path.exists(target):
            click.echo(as_link(target, Color.BRIGHT_GREEN))
        else:
            click.echo(
                f"{as_link(target, Color.RED)}" " (broken, links to non-existent file)."
            )
    except FileNotFoundError:
        click.echo(f'{as_link("null", Color.YELLOW)} (does not link to any file).')



"""choose between local, existing, .env files."""
import os
from typing import Optional

import click
import questionary

from chenv import settings
from chenv.cli import cli
from chenv.console import fatal
from chenv.models.output import Target


@cli.command("local", help="choose between local, existing, .env files")
@click.argument("filename", required=False, metavar="filename")
def collect(filename: Optional[str]) -> Target:
    """Choose between local, existing, .env files."""
    if filename:
        filename = filename.replace(settings.PREFIX, "")
        if os.path.exists(settings.filename_from_template(filename)):
            return Target(file_suffix=filename)

    env_files = sorted(
        filename.replace(settings.PREFIX, "")
        for filename in os.listdir(".")
        if filename.startswith(settings.PREFIX)
    )

    if not env_files:
        fatal(__name__, "No local options available.", 2)

    click.echo(
        f"""Local options: {", ".join(click.style(env_file, fg="magenta")
         for env_file in env_files)}"""
    )
    env_file = questionary.autocomplete("Choose file:", choices=env_files).ask()
    return Target(file_suffix=env_file)



"""declares outputs used to generate `.env`-style files."""
from dataclasses import dataclass
import os
import shlex

from chenv import settings


def _escape(value: str) -> str:
    """Escape variable value inline."""
    return shlex.quote(value).replace(os.linesep, r"\n")


@dataclass(frozen=True)
class Target:
    """Declares a target file to use as the `.env` file."""

    file_suffix: str

    @property
    def filename(self) -> str:  # noqa: ANN101
        """Generate a filename from this `Target` (by `file_suffix`)."""
        return settings.filename_from_template(self.file_suffix)


@dataclass(frozen=True)
class Output(Target):
    """Declares a target file, and body, to use as the `.env` file."""

    variables: dict

    @property
    def body(self) -> str:  # noqa: ANN101
        """Generate a file body from the `self.variables` dict."""
        return os.linesep.join(
            f"{key}={_escape(value)}" for key, value in self.variables.items()
        )



"""chenv. modern local environment management."""
from chenv import cli, settings

settings.load()


def main() -> None:
    """Main function / entry point for the application."""
    from chenv import subcommands  # noqa: F401

    cli.cli()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['chenv',
 'chenv.models',
 'chenv.subcommands',
 'chenv.subcommands.blank',
 'chenv.subcommands.heroku',
 'chenv.subcommands.local',
 'chenv.subcommands.which']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'desert>=2020.11.18,<2021.0.0',
 'httpx>=0.16.1,<0.17.0',
 'marshmallow>=3.10.0,<4.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'questionary>=1.9.0,<2.0.0',
 'toolz>=0.11.1,<0.12.0']

entry_points = \
{'console_scripts': ['chenv = chenv.main:main']}

setup_kwargs = {
    'name': 'chenv',
    'version': '0.3.1.dev1611763613',
    'description': 'modern local environment management',
    'long_description': ".. code-block:: text\n\n          _\n      ___| |__   ___ _ ____   __\n     / __| '_ \\ / _ | '_ \\ \\ / /\n    | (__| | | |  __| | | \\ V /\n     \\___|_| |_|\\___|_| |_|\\_/ . modern local environment management\n\n|Status| |PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |Status| image:: https://badgen.net/badge/status/alpha/d8624d\n   :target: https://badgen.net/badge/status/alpha/d8624d\n   :alt: Project Status\n.. |PyPI| image:: https://img.shields.io/pypi/v/chenv.svg\n   :target: https://pypi.org/project/chenv/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/chenv\n   :target: https://pypi.org/project/chenv\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/chenv\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/chenv/latest.svg?label=Read%20the%20Docs\n   :target: https://chenv.readthedocs.io/\n   :alt: Read the documentation at https://chenv.readthedocs.io/\n.. |Tests| image:: https://github.com/jonathan-shemer/chenv/workflows/Tests/badge.svg\n   :target: https://github.com/jonathan-shemer/chenv/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/jonathan-shemer/chenv/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/jonathan-shemer/chenv\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nThis command-line interface creates and manages local `.env` files from various sources.\n\nCoupled with `python-dotenv <https://pypi.org/project/python-dotenv/>`_ for python,\nor `dotenv <https://www.npmjs.com/package/dotenv/>`_ for node.js development,\nit provides better, more consistent environment variable management and developement.\n\nInstallation\n------------\n\nTo install `chenv`,\nrun this command in your terminal:\n\n.. code-block:: shell\n\n   $ pip install --user chenv\n\nAlso make sure that your :code:`$PATH` includes :code:`$HOME/.local/bin`.\nIf not, add this line to your :code:`.bashrc` / :code:`.zshrc`:\n\n.. code-block:: shell\n\n   export PATH=$HOME/.local/bin:$PATH;\n\nUsage\n-----\n\n`chenv`'s usage looks like:\n\n.. code-block:: shell\n\n   $ chenv COMMAND [ARGS]\n\nCommands currently include:\n\n=====\nblank\n=====\n\n   Choose to set `.env` as a new, blank, `.env.blank` file.\n\n======\nheroku\n======\n\n   Choose to set `.env` from a remote heroku app config-vars, as `.env.[app-name]`.\n\n   - -t <team>, --team <team>\n       Pre-fill team name\n\n   - -a <app>, --app <app>\n      Pre-fill app name\n\n=====\nlocal\n=====\n\n   Choose to set `.env` from a local, pre-exsiting `.env.*` file.\n\n   - filename\n      Pre-fill file-suffix name\n\nProject Configurations\n----------------------\n\n`chenv` also provides two file types that manipulate the output of new `.env.*` files being set.\n\n==========\n.envignore\n==========\n\n   Specifies intentionally unwanted environment-variables.\n   Each line in a envignore file specifies a pattern.\n\n   When deciding whether to ignore an environment variable, `chenv` checks it's key against the list of patterns described in this file.\n\n   :Pattern:\n      `.envignore` uses the unix filename pattern matching, similar to `.gitignore`'s, and as specified at https://docs.python.org/3/library/fnmatch.html\n\n=========\n.envmerge\n=========\n\n   Sepecifies environment variables to merge / override after any input is chosen. This provides consistency to preffered settings such as the `logging-level`, or `NODE_ENV` for local development usage in node.js.\n",
    'author': 'Jonathan Shemer',
    'author_email': 'i@jonathanshemer.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonathan-shemer/chenv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
